﻿using LINQPad;
using LINQPad.Extensibility.DataContext;
using Microsoft.CSharp;
using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Xml;

namespace DataContextDriverDemo.DynamicDemo
{
	public class DynamicDemoDriver : DynamicDataContextDriver
	{
		static DynamicDemoDriver ()
		{
			// Uncomment the following code to attach to Visual Studio's debugger when an exception is thrown.
			//AppDomain.CurrentDomain.FirstChanceException += (sender, args) =>
			//{
			//	if (args.Exception.StackTrace.Contains (typeof (DynamicDemoDriver).Namespace))
			//		Debugger.Launch ();
			//};
		}

		public override string Name => "Dynamic Demo Driver";

		public override string Author => "Joe Albahari";

		public override string GetConnectionDescription (IConnectionInfo cxInfo)
			=> "Dynamic demo - " + new DynamicDemoProperties (cxInfo).RandomPropertyCount + " properties";

		public override bool ShowConnectionDialog (IConnectionInfo cxInfo, ConnectionDialogOptions dialogOptions) 
			=> new ConnectionDialog (cxInfo).ShowDialog () == true;

		public override List<ExplorerItem> GetSchemaAndBuildAssembly (
			IConnectionInfo cxInfo, AssemblyName assemblyToBuild, ref string nameSpace, ref string typeName)
		{
			// Retrieve the property that the user entered in the connection dialog.
			int propertyCount = new DynamicDemoProperties (cxInfo).RandomPropertyCount;

			// Generate C# code for each of the properties we want to generate:
			var randomProperties = Enumerable
				.Range (0, propertyCount)
				.Select (i => $"	public string Prop{i} => Guid.NewGuid().ToString();");

			string allRandomProps = string.Join ("\r\n", randomProperties);

			// Our source will consist of a single queryable object inside a typed data context.

			string source = @"using System;
using System.Collections.Generic;
using System.Linq;

namespace " + nameSpace + @"
{
    // The main typed data class. The user's queries subclass this, so they have easy access to all its members.
	public class " + typeName + @"
	{
        // We would normally have more than one property here.
		public IEnumerable<MyRandomData> RandomData => Enumerable.Range (1, 100).Select (i => new MyRandomData());
	}	

	// This is a typed class that the user will be querying. They'd typically be more than one of these, too.
	public class MyRandomData    
	{
" + allRandomProps + @"
	}
}";
			Compile (source, assemblyToBuild.CodeBase);

			// We need to tell LINQPad what to display in the TreeView on the left (Schema Explorer):
			var schema =
				new ExplorerItem ("RandomData", ExplorerItemKind.QueryableObject, ExplorerIcon.Table)
				{
					IsEnumerable = true,
					DragText = "RandomData",					
					Children = 
						Enumerable.Range (0, propertyCount).Select (i =>
							new ExplorerItem ("Prop" + i, ExplorerItemKind.Property, ExplorerIcon.Column)
						).ToList ()
				};

			return new[] { schema }.ToList();
		}

		static void Compile (string cSharpSourceCode, string outputFile)
		{
			string[] assembliesToReference =
#if NETCORE
			// GetCoreFxReferenceAssemblies is helper method that returns the full set of .NET Core reference assemblies.
			// (There are more than 100 of them.)
			GetCoreFxReferenceAssemblies ();
#else
			// .NET Framework - here's how to get the basic Framework assemblies:
			new[]
			{
				typeof (int).Assembly.Location,            // mscorlib
				typeof (Uri).Assembly.Location,            // System
				typeof (XmlConvert).Assembly.Location,     // System.Xml
				typeof (Enumerable).Assembly.Location,     // System.Core
				typeof (DataSet).Assembly.Location         // System.Data
			};
#endif
			// CompileSource is a static helper method to compile C# source code using LINQPad's built-in Roslyn libraries.
			// If you prefer, you can add a NuGet reference to the Roslyn libraries and use them directly.
			var compileResult = CompileSource (new CompilationInput
			{
				FilePathsToReference = assembliesToReference,
				OutputPath = outputFile,
				SourceCode = new[] { cSharpSourceCode }
			});

			if (compileResult.Errors.Length > 0)
				throw new Exception ("Cannot compile typed context: " + compileResult.Errors[0]);
		}
	}
}
